--                XX-bit to Hi-/TrueColor convertion routines                       
-- The RGB-to-HiColor convertion formulas were "borrowed" from Pete Eberlein's Neil.e 

include asm.e
include ports.e

global constant SYSTEM_PALETTE=12321

function flip(sequence s)
sequence s2
 s2=s
 for i=1 to length(s) do
  s2[length(s)-(i-1)]=s[i]
 end for
 return s2
end function

procedure poke_word(atom dest,object word)
 if sequence(word) then
  for i=1 to length(word) do
   poke(dest+((i-1)*2),and_bits(word[i],#00FF))
   poke(dest+((i-1)*2)+1,floor(and_bits(word[i],#FF00)/256))
  end for
 else
  poke(dest,and_bits(word,#00FF))
  poke(dest+1,floor(and_bits(word,#FF00)/256))
 end if
end procedure

global function makecol15(sequence rgb)
integer col
 col=floor(rgb[3]/8)+and_bits(rgb[2]*4,#E0)
 col=col+((floor(rgb[2]/64)+and_bits(rgb[1]/2,#7C))*256)
 return col
end function


global function makecol16(sequence rgb)
integer col
 col=floor(rgb[3]/8)+and_bits(rgb[2]*8,#E0)
 col=col+((floor(rgb[2]/32)+and_bits(rgb[1],#F8))*256)
 return col
end function

global function makecol24(sequence rgb)
integer col
 col=rgb[3]+(rgb[2]*256)+(rgb[1]*65536)
 return col
end function


global function convert_image(sequence img,integer dest_bpp,integer source_bpp)
atom mem
integer r,g,b
if dest_bpp=16 then
 mem=allocate(img[2]*img[3]*2)
 if mem=0 then
  return {-1}&img
 end if
 if source_bpp=24 then
  for i=1 to img[3] do
   for j=1 to img[2] do
    poke_word(mem+((i-1)*img[2]*2)+((j-1)*2),makecol16(flip(peek({img[1]+((i-1)*img[2]*3)+((j-1)*3),3}))))
   end for
  end for
  free(img[1])
 elsif source_bpp=8 then	--img[4] must be a palette
  for i=1 to img[3] do
   for j=1 to img[2] do
    poke_word(mem+((i-1)*img[2]*2)+((j-1)*2),makecol16(img[4][peek(img[1]+((i-1)*img[2])+(j-1))+1]))
   end for
  end for
 end if
elsif dest_bpp=15 then
 mem=allocate(img[2]*img[3]*2)
 if mem=0 then
  return {-1}&img
 end if
 if source_bpp=24 then
  for i=1 to img[3] do
   for j=1 to img[2] do
    poke_word(mem+((i-1)*img[2]*2)+((j-1)*2),makecol15(flip(peek({img[1]+((i-1)*img[2]*3)+((j-1)*3),3}))))
   end for
  end for
  free(img[1])
 elsif source_bpp=8 then	--img[4] must be a palette
  for i=1 to img[3] do
   for j=1 to img[2] do
    poke_word(mem+((i-1)*img[2]*2)+((j-1)*2),makecol15(img[4][peek(img[1]+((i-1)*img[2])+(j-1))+1]))
   end for
  end for
 elsif source_bpp=16 then
  for i=1 to img[3] do
   for j=1 to img[2] do
    b=peek(img[1]+((i-1)*img[2]*2)+((j-1)*2))
    b=b+(peek(img[1]+((i-1)*img[2]*2)+((j-1)*2)+1)*256)
    r=floor(and_bits(b,#F800)/2048)
    g=floor(and_bits(b,#07E0)/32)
    b=and_bits(b,#001F)
    poke_word(mem+((i-1)*img[2]*2)+((j-1)*2),b+(floor(g*(31/63))*32)+(r*1024))
   end for
  end for
 end if
elsif dest_bpp=24 then
 mem=allocate(img[2]*img[3]*3)
 if mem=0 then
  return {-1}&img
 end if
 if source_bpp=8 then
  for i=1 to img[3] do
   for j=1 to img[2] do
    poke(mem+((i-1)*img[2]*3)+((j-1)*3),img[4][peek(img[1]+((i-1)*img[2])+(j-1))+1][3])
    poke(mem+((i-1)*img[2]*3)+((j-1)*3)+1,img[4][peek(img[1]+((i-1)*img[2])+(j-1))+1][2])
    poke(mem+((i-1)*img[2]*3)+((j-1)*3)+2,img[4][peek(img[1]+((i-1)*img[2])+(j-1))+1][1])
   end for
  end for  
 elsif source_bpp=16 then
  for i=1 to img[3] do
   for j=1 to img[2] do
    b=peek(img[1]+((i-1)*img[2]*2)+((j-1)*2))
    b=b+(peek(img[1]+((i-1)*img[2]*2)+((j-1)*2)+1)*256)
    r=floor(and_bits(b,#F800)/2048)
    g=floor(and_bits(b,#07E0)/32)
    b=and_bits(b,#001F)
    poke(mem+((i-1)*img[2]*3)+((j-1)*3),floor(b*(255/31)))
    poke(mem+((i-1)*img[2]*3)+((j-1)*3)+1,floor(g*(255/63)))
    poke(mem+((i-1)*img[2]*3)+((j-1)*3)+2,floor(r*(255/31)))
   end for
  end for
 end if
else
 return {-1}&img
end if
return {mem,img[2],img[3]}
end function


function get_pal()
sequence ret
 ret=repeat({0,0,0},256)
 for i=1 to 256 do
  Output((i-1),#3C7)
  ret[i][1]=Input(#3C9)
  ret[i][2]=Input(#3C9)
  ret[i][3]=Input(#3C9)
 end for
 return ret
end function


constant match_pal=allocate(768*4),
         match_data=allocate(32),
         match_REFR=0,match_REFG=4,match_REFB=8,
         match_DIFR=12,match_DIFG=16,match_DIFB=20,
         match_MATCH=24,match_MINV=28

constant match_fast=get_asm(
"pusha "&
"xor esi,esi "&
"xor edi,edi "&
"mf_@loop: "&
"xor eax,eax xor ebx,ebx "&
"mov eax,[refR] "&	
"sub eax,[candR + esi] "&
"mov edx,eax "& --\
"sar edx,15 "&	-- |	
"xor eax,edx "&	-- | calculate absolute value
"sub eax,edx "& --/
"mov [difRsqr],eax "&
"mov eax,[refG] "&	
"sub eax,[candG + esi] "&
"mov edx,eax "& --
"sar edx,15 "&
"xor eax,edx "&
"sub eax,edx "& --
"mov [difGsqr],eax "&
"mov eax,[refB] "&	
"sub eax,[candB + esi] "&
"mov edx,eax "& --
"sar edx,15 "&
"xor eax,edx "&
"sub eax,edx "& --
"add eax,[difGsqr] "&
"add eax,[difRsqr] "&
"mov ebx,eax "&
"imul ebx "&
"mov [difBsqr],eax "&
"cmp eax,0 "&
"jne mf_@not_perfect_match "&
"mov [match],edi "&
"popa ret "&
"mf_@not_perfect_match: "&
"cmp eax,[min_vector] "&
"jge mf_@no_min "&
"mov [min_vector],eax "&
"mov [match],edi "&
"mf_@no_min: "&
"add esi,12 "&
"inc edi "&
"cmp edi,255 "&
"jg mf_@done "&
"mov ebx,mf_again "&
"jmp near ebx "&
"mf_@done: "&
"popa ret")
resolve_param("candR",match_pal)
resolve_param("candG",match_pal+4)
resolve_param("candB",match_pal+8)
resolve_param("refR",match_data+match_REFR)
resolve_param("refG",match_data+match_REFG)
resolve_param("refB",match_data+match_REFB)
resolve_param("difRsqr",match_data+match_DIFR)
resolve_param("difGsqr",match_data+match_DIFG)
resolve_param("difBsqr",match_data+match_DIFB)
resolve_param("match",match_data+match_MATCH)
resolve_param("min_vector",match_data+match_MINV)
resolve_param("mf_again",match_fast+get_label("mf_@loop"))


function find_rgb(sequence rgb)
 poke4(match_data+match_REFR,rgb[1])
 poke4(match_data+match_REFG,rgb[2])
 poke4(match_data+match_REFB,rgb[3])
 poke4(match_data+match_MATCH,0)
 poke4(match_data+match_MINV,13107200)
 call(match_fast)
 return peek4s(match_data+match_MATCH)
end function



constant syspal={
{0,0,0},{0,0,42},{0,42,0},{0,42,42},{42,0,0},{42,0,42},{42,21,0},{42,42,42},{21,21,21},{21,21,63},{21,63,21},
{21,63,63},{63,21,21},{63,21,63},{63,63,21},{63,63,63},{0,0,0},{5,5,5},{8,8,8},{11,11,11},{14,14,14},{17,17,17},
{20,20,20},{24,24,24},{28,28,28},{32,32,32},{36,36,36},{40,40,40},{45,45,45},{50,50,50},{56,56,56},{63,63,63},{0,0,63},
{16,0,63},{31,0,63},{47,0,63},{63,0,63},{63,0,47},{63,0,31},{63,0,16},{63,0,0},{63,16,0},{63,31,0},{63,47,0},{63,63,0},
{47,63,0},{31,63,0},{16,63,0},{0,63,0},{0,63,16},{0,63,31},{0,63,47},{0,63,63},{0,47,63},{0,31,63},{0,16,63},{31,31,63},
{39,31,63},{47,31,63},{55,31,63},{63,31,63},{63,31,55},{63,31,47},{63,31,39},{63,31,31},{63,39,31},{63,47,31},
{63,55,31},{63,63,31},{55,63,31},{47,63,31},{39,63,31},{31,63,31},{31,63,39},{31,63,47},{31,63,55},{31,63,63},
{31,55,63},{31,47,63},{31,39,63},{45,45,63},{49,45,63},{54,45,63},{58,45,63},{63,45,63},{63,45,58},{63,45,54},
{63,45,49},{63,45,45},{63,49,45},{63,54,45},{63,58,45},{63,63,45},{58,63,45},{54,63,45},{49,63,45},{45,63,45},
{45,63,49},{45,63,54},{45,63,58},{45,63,63},{45,58,63},{45,54,63},{45,49,63},{0,0,28},{7,0,28},{14,0,28},{21,0,28},
{28,0,28},{28,0,21},{28,0,14},{28,0,7},{28,0,0},{28,7,0},{28,14,0},{28,21,0},{28,28,0},{21,28,0},{14,28,0},{7,28,0},
{0,28,0},{0,28,7},{0,28,14},{0,28,21},{0,28,28},{0,21,28},{0,14,28},{0,7,28},{14,14,28},{17,14,28},{21,14,28},{24,14,28},
{28,14,28},{28,14,24},{28,14,21},{28,14,17},{28,14,14},{28,17,14},{28,21,14},{28,24,14},{28,28,14},{24,28,14},{21,28,14},
{17,28,14},{14,28,14},{14,28,17},{14,28,21},{14,28,24},{14,28,28},{14,24,28},{14,21,28},{14,17,28},{20,20,28},{22,20,28},
{24,20,28},{26,20,28},{28,20,28},{28,20,26},{28,20,24},{28,20,22},{28,20,20},{28,22,20},{28,24,20},{28,26,20},{28,28,20},
{26,28,20},{24,28,20},{22,28,20},{20,28,20},{20,28,22},{20,28,24},{20,28,26},{20,28,28},{20,26,28},{20,24,28},{20,22,28},
{0,0,16},{4,0,16},{8,0,16},{12,0,16},{16,0,16},{16,0,12},{16,0,8},{16,0,4},{16,0,0},{16,4,0},{16,8,0},{16,12,0},{16,16,0},
{12,16,0},{8,16,0},{4,16,0},{0,16,0},{0,16,4},{0,16,8},{0,16,12},{0,16,16},{0,12,16},{0,8,16},{0,4,16},{8,8,16},{10,8,16},
{12,8,16},{14,8,16},{16,8,16},{16,8,14},{16,8,12},{16,8,10},{16,8,8},{16,10,8},{16,12,8},{16,14,8},{16,16,8},{14,16,8},
{12,16,8},{10,16,8},{8,16,8},{8,16,10},{8,16,12},{8,16,14},{8,16,16},{8,14,16},{8,12,16},{8,10,16},{11,11,16},{12,11,16},
{13,11,16},{15,11,16},{16,11,16},{16,11,15},{16,11,13},{16,11,12},{16,11,11},{16,12,11},{16,13,11},{16,15,11},{16,16,11},
{15,16,11},{13,16,11},{12,16,11},{11,16,11},{11,16,12},{11,16,13},{11,16,15},{11,16,16},{11,15,16},{11,13,16},{11,12,16},
{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}}


global procedure makecol_palette(object mkcol_pal)
sequence junk
 if (integer(mkcol_pal)) then
  if mkcol_pal=SYSTEM_PALETTE then
   junk=syspal
  end if
 else
  junk=mkcol_pal
 end if
 for i=0 to 255 do
  poke4(match_pal+(i*3*4),junk[i+1])
 end for
end procedure


global function makecol8(sequence rgb)
 rgb=floor(rgb/4)
 return find_rgb(rgb)
end function

global function makecol(integer bpp,sequence rgb)
 if bpp=15 then
  return makecol15(rgb)
 elsif bpp=16 then
  return makecol16(rgb)
 elsif bpp=24 then
  return makecol24(rgb)
 elsif bpp=8 then
  return makecol8(rgb)
 end if
end function

global function create_mixtable(sequence pal1,sequence pal2,object pal_used)
sequence mixpal
atom mixtb
  if (integer(pal_used)) then
    if pal_used=SYSTEM_PALETTE then
      mixpal = syspal
    end if
  else
    mixpal = pal_used
  end if
  mixtb = allocate(65536)
  if mixtb<=0 then
    return 0
  end if
  for i = 0 to 255 do
    poke4(match_pal+(i*3*4),mixpal[i+1])
  end for
  for i=0 to 255 do
    for j=0 to 255 do
      poke(mixtb+(i*256)+j,find_rgb(floor((pal1[i+1]+pal2[j+1])/2)))
    end for
  end for
  return mixtb
end function




